Merge branch 'maint'
[paraslash.git] / ipc.c
diff --git a/ipc.c b/ipc.c
index 9ae2b3a09d1e1d98617e4502c16f7af2f0f54742..c64cc89c6d8803416a9686890b09592c97a9272e 100644 (file)
--- a/ipc.c
+++ b/ipc.c
@@ -1,35 +1,70 @@
+/*
+ * Copyright (C) 2006-2009 Andre Noll <maan@systemlinux.org>
+ *
+ * Licensed under the GPL v2. For licencing details see COPYING.
+ */
+
+/** \file ipc.c Inter-process communication and shared memory helpers. */
+
 #include "para.h"
 #include "error.h"
 #include "ipc.h"
 #include <sys/ipc.h>
 #include <sys/shm.h>
+#include <sys/sem.h>
 
-
+/**
+ * Define a new mutex.
+ *
+ * \return The identifier for the new mutex on success, a negative error code
+ * on errors.
+ *
+ * \sa semget(2).
+ */
 int mutex_new(void)
 {
        int ret = semget(IPC_PRIVATE, 1, IPC_CREAT | 0666);
-       return ret < 0?  -E_SEM_GET : ret;
+       return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : ret;
 }
 
-int mutex_remove(int id)
+/**
+ * Destroy a mutex.
+ *
+ * \param id The identifier of the mutex to be destroyed.
+ *
+ * \return Standard.
+ *
+ * \sa semctl(2)
+ */
+int mutex_destroy(int id)
 {
        int ret = semctl(id, 0, IPC_RMID);
-       return ret < 0? -E_SEM_REMOVE : 1;
+       return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : 1;
 }
 
 static void para_semop(int id, struct sembuf *sops, int num)
 {
-       if (semop(id, sops, num) >= 0)
+       do {
+               if (semop(id, sops, num) >= 0)
+                       return;
+       } while (errno == EINTR);
+       if (errno == EIDRM) {
+               PARA_CRIT_LOG("semaphore set %d was removed\n", id);
                return;
-       PARA_WARNING_LOG("semop failed (%s), retrying\n", strerror(errno));
-       while (semop(id, sops, num) < 0)
-               ; /* nothing */
+       }
+       PARA_EMERG_LOG("fatal semop error %s: pid %d\n", strerror(errno),
+               (int)getpid());
+       exit(EXIT_FAILURE);
 }
 
 /**
- * lock the given mutex
+ * Lock the given mutex.
+ *
+ * \param id The identifier of the shared memory area to lock.
  *
- * \sa semop(2), struct misc_meta_data
+ * This function either succeeds or aborts.
+ *
+ * \sa semop(2), struct misc_meta_data.
  */
 void mutex_lock(int id)
 {
@@ -49,9 +84,13 @@ void mutex_lock(int id)
 }
 
 /**
- * unlock a mutex
+ * Unlock a mutex.
+ *
+ * \param id The identifier of the mutex.
  *
- * \sa semop(2), struct misc_meta_data
+ * This function either succeeds or aborts.
+ *
+ * \sa semop(2), struct misc_meta_data.
  */
 void mutex_unlock(int id)
 {
@@ -66,40 +105,69 @@ void mutex_unlock(int id)
 }
 
 /**
* create a new shared memory area of given size
* 
* \sa shmget(2)
  */
 int shm_new(size_t size)
 {
        int ret = shmget(IPC_PRIVATE, size, IPC_CREAT | IPC_EXCL | 0600);
-       return ret < 0 ? -E_SHM_GET : ret;
+       return ret < 0 ? -ERRNO_TO_PARA_ERROR(errno) : ret;
 }
 
 /**
- * destroy the given shared memory area
- * \sa shmctl(2)
- **/
+ * Destroy the given shared memory area.
+ *
+ * \param id The shared memory identifier.
+ *
+ * \return The return value of the underlying shmctl() call on success,
+ * a negative error code on errors.
+ *
+ * \sa shmctl(2).
+ */
 int shm_destroy(int id)
 {
        struct shmid_ds shm_desc;
        int ret = shmctl(id, IPC_RMID, &shm_desc);
-       return ret < 0? -E_SHM_DESTROY : ret;
+       return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : ret;
 }
 
 /**
- * attach a shared memory area
+ * Attach a shared memory segment.
  *
- * \sa semop(2)
+ * \param id The identifier of the shared memory segment to attach.
+ * \param mode Either ATTACH_RO (read only) or ATTACH_RW (read/write).
+ * \param result Points to the attached area just attached on success.
+ *
+ * \return Standard.
+ *
+ * \sa shmat(2).
  */
-void *shm_attach(int id, enum shm_attach_mode mode)
+int shm_attach(int id, enum shm_attach_mode mode, void **result)
 {
        if (mode == ATTACH_RW)
-               return shmat(id, NULL, 0);
-       return shmat(id, NULL, SHM_RDONLY);
+               *result = shmat(id, NULL, 0);
+       else
+               *result = shmat(id, NULL, SHM_RDONLY);
+       return *result == (void *) -1? -ERRNO_TO_PARA_ERROR(errno) : 1;
 }
+
+/**
+ * Detach a shared memory segment.
+ *
+ * \param addr The address of the attached segment.
+ *
+ * \return Standard.
+ *
+ * \sa shmdt(2).
+ */
 int shm_detach(void *addr)
 {
        int ret = shmdt(addr);
-       return ret < 0? -E_SHM_DETACH : 1;
+       return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : 1;
 }
@@ -66,40 +105,69 @@ void mutex_unlock(int id)
 
 /**
- * create a new shared memory area of given size
- * 
- * \sa shmget(2)
+ * Create a new shared memory area of given size.
+ *
+ * \param size The size of the shared memory area to create.
+ *
+ * \return The id of the shared memory array on success, a negative error
+ * code on errors.
+ *
+ * \sa shmget(2).
  */
 int shm_new(size_t size)